from abc import ABC, abstractmethod
import river as rv


class Model(ABC):
    _model: rv.base.regressor

    def __init__(self, **kwargs):
        pass

    def get_model(self):
        return self._model



from collections import OrderedDict
from numbers import Number
from typing import List, Callable, Union
import numpy as np

from movado.controller import Controller
from movado.estimator import Estimator
from functools import wraps
from movado.distance_controller import DistanceController
from movado.hoeffding_adaptive_tree_model import HoeffdingAdaptiveTreeModel
from movado.chained_estimator import ChainedEstimator
import movado.movado_static


# This imports are unused but populate the global symbol table for the globals() call
# noinspection PyUnresolvedReferences
from movado.mab_controller import MabController  # pylint: disable=unused-import

# noinspection PyUnresolvedReferences
from movado.voting_controller import VotingController  # pylint: disable=unused-import

# noinspection PyUnresolvedReferences
from movado.shadow_controller import ShadowController

# noinspection PyUnresolvedReferences
from movado.kernel_regression_model import (
    KernelRegressionModel,
)  # pylint: disable=unused-import


def approximate(
    **kwargs,
) -> Callable[
    [Callable[[List[float]], List[float]]], Callable[[List[float]], List[float]]
]:
    def approximate_decorator(
        func: Callable[[List[float]], List[float]]
    ) -> Callable[[List[float]], List[float]]:
        controller: Controller
        estimator: Estimator
        is_first_call: bool = True
        outputs: int = -1

        @wraps(func)
        def wrapper(*wrapper_args) -> Union[List[float], float]:
            # args must either contain the point to be analyzed
            # or it must contain self (for class methods) and then the point
            nonlocal is_first_call
            nonlocal controller
            nonlocal estimator
            nonlocal outputs

            if len(wrapper_args) < 1 or len(wrapper_args) > 2:
                raise Exception(
                    "The decorated function must have a single input which is a list of numbers, "
                    + "self may also be present for non-static class methods, in such case self is "
                    + "expected to be the first argument followed by the point and nothing"
                )
            if len(wrapper_args) == 1:
                point = wrapper_args[0]
            else:
                point = wrapper_args[1]
            if is_first_call:
                is_first_call = False
                selected_model = globals().get(str(kwargs.get("estimator")) + "Model")
                model = (
                    HoeffdingAdaptiveTreeModel if not selected_model else selected_model
                )
                # TODO add multi-output non-chained estimators
                outputs = kwargs.get("outputs")
                voters: int = kwargs.get("voters")
                if not voters:
                    voters = 750
                if not outputs:
                    raise Exception(
                        "Please specify outputs as a kwarg as the number of targets of the fitness function"
                    )
                estimator = ChainedEstimator(
                    model(
                        **{
                            k[k.find("_") + 1 :]: v
                            for k, v in kwargs.items()
                            if k[: k.find("_")] == "estimator"
                        }
                    ),
                    outputs,
                )
                selected_controller = globals().get(
                    str(kwargs.get("controller")) + "Controller"
                )
                controller = (
                    DistanceController
                    if not selected_controller
                    else selected_controller
                )
                if selected_controller is MabController:
                    params = OrderedDict(
                        {
                            "cover": list(range(3, 20)),
                            "mab_weight_epsilon": [i / 100 for i in range(1, 41)],
                            "mab_weight_bandwidth": list(range(3, 7)),
                        }
                    )
                else:
                    params = OrderedDict(
                        {
                            "nth_nearest": list(range(1, 6)),
                            "mab_epsilon": [i / 100 for i in range(1, 41)],
                            "mab_bandwidth": list(range(3, 7)),
                            "mab_weight_epsilon": [i / 100 for i in range(1, 41)],
                            "mab_weight_bandwidth": list(range(3, 7)),
                        }
                    )
                available_models = int(np.prod([len(v) for v in params.values()]))
                voters = voters if voters <= available_models else available_models
                if voters <= 1:
                    controller = controller(
                        exact_fitness=func,
                        estimator=estimator,
                        self_exact=None if len(wrapper_args) == 1 else wrapper_args[0],
                        **{
                            k[k.find("_") + 1 :]: v
                            for k, v in kwargs.items()
                            if k[: k.find("_")] == "controller"
                        },
                    )
                else:
                    controller = VotingController(
                        controller,
                        estimator,
                        func,
                        voters,
                        params,
                        self_exact=None if len(wrapper_args) == 1 else wrapper_args[0],
                        **{
                            k[k.find("_") + 1 :]: v
                            for k, v in kwargs.items()
                            if k[: k.find("_")] == "controller"
                        },
                    )
            if outputs == 1:
                return list(controller.compute_objective(point))[0]
            return list(controller.compute_objective(point))

        if kwargs.get("disabled"):
            return func
        return wrapper

    return approximate_decorator



from typing import List, Tuple

from vowpalwabbit import pyvw

from movado.mab_handler import MabHandler


class MabHandlerCATS(MabHandler):
    def __init__(
        self,
        bandwidth: int = 1,
        epsilon: float = 0.2,
        debug=False,
        controller_params: dict = None,
        debug_path: str = "mab",
        skip_debug_initialization: bool = False,
    ):
        """
        CATS parameters interpretation:
        - the range of outputs is (min_value, max_value), in our case (0, 100)
        - mab_actions determines the number of buckets (discretization) we apply to the range,
          in our case range is actions is 100 thus we have 100 buckets of length 1 ((0, 1], [1, 2], ..., [99, 100))
        - bandwidth determines the density function of the outputs for each bucket outputs
          will be more dense around [bucket_center+bandwidth, bucket_center-bandwidth]
          where bucket_center = (bucket_min + bucket_max)/2. In our case bandwidth defaults to 1.
        """
        super(MabHandlerCATS, self).__init__(
            debug,
            debug_path,
            controller_params=controller_params,
            skip_debug_initialization=skip_debug_initialization,
        )

        self._sample_prefix = "ca "
        if bandwidth < 0:
            raise Exception(
                "Invalid bandwidth value: "
                + str(bandwidth)
                + " it must be greater or equal than 1"
            )
        if epsilon < 0 or epsilon > 1:
            raise Exception(
                "Invalid epsilon value: " + str(epsilon) + "it must be between 0 and 1"
            )
        mab_actions: int = 100
        self._mab: pyvw.vw = pyvw.vw(
            "--cats "
            + str(mab_actions)
            + "  --bandwidth "
            + str(bandwidth)
            + " --min_value 0 --max_value 100 --chain_hash --coin --epsilon "
            + str(epsilon)
            + " --quiet"
            + " --random_seed 0"
        )

    def predict(self, context: List[float]) -> float:
        context_str: str = "| "
        for feature in context:
            context_str += str(feature) + " "
        context_str.strip()
        action: Tuple[float, float] = self._mab.predict(context_str)
        self._last_action = action[0]
        self._last_predict_probability = action[1]
        return action[0]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['movado']

package_data = \
{'': ['*']}

install_requires = \
['joblib>=1.0,<2.0',
 'numpy>=1.19,<2.0',
 'plotly>=4.14,<5.0',
 'river==0.14',
 'scikit-learn>=0.24,<0.25',
 'scipy>=1.5,<2.0',
 'vowpalwabbit>=8,<9']

setup_kwargs = {
    'name': 'movado',
    'version': '0.6.5',
    'description': 'Approximation utility for expensive fitness functions',
    'long_description': 'None',
    'author': 'Daniele Paletti',
    'author_email': 'danielepaletti98@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
